#!/usr/bin/env python

# Script authored by FM Ytreberg <fmytreberg@gmail.com>.
# Last modified Jan 2015.
#
# For usage information run with '-h' flag.
#
# Requires foldx software (http://foldx.crg.es/).

#
# IMPORTS
#

import argparse
import datetime
import glob
import os
import shutil
import socket
import subprocess
import sys

# May need to change variables below #
foldx = '/home/marty/apps/foldx'
pbs_foldx = '/mnt/home/mytrebe/apps/foldx'
######################################

#
# CONSTANTS
#

SELF_MUT_DG_MAX = 0.5
RESI_DICT = {
        'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
        'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N', 
        'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 
        'ALA': 'A', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M' }

# Note that changing to <pdb_hydrogens>true causes foldx to output hydrogens
# in the PDB file, but the numerical free energy results are identical to
# <pdb_hydrogens>false.
FX_REPAIR = '''<TITLE>FOLDX_runscript;
<JOBSTART>#;
    <PDBS>#;
    <BATCH>listNUM.txt;
    <COMMANDS>FOLDX_commandfile;
        <RepairPDB>#;
    <END>#;
    <OPTIONS>FOLDX_optionfile;
        <Temperature>TEMP;
        <R>#;
        <pH>7;
        <IonStrength>0.050;
        <water>-IGNORE;
        <metal>-IGNORE;
        <VdWDesign>2;
        <OutPDB>true;
        <pdb_hydrogens>false;
    <END>#;
<JOBEND>#;
<ENDFILE>#;
'''

FX_BUILD = '''<TITLE>FOLDX_runscript;
<JOBSTART>#;
    <PDBS>#;
    <BATCH>listNUM.txt;
    <COMMANDS>FOLDX_commandfile;
        <BuildModel>#,individual_list.txt;
    <END>#;
    <OPTIONS>FOLDX_optionfile;
        <Temperature>TEMP;
        <R>#;
        <pH>7;
        <IonStrength>0.050;
        <water>-IGNORE;
        <metal>-IGNORE;
        <VdWDesign>2;
        <OutPDB>true;
        <pdb_hydrogens>false;
    <END>#;
<JOBEND>#;
<ENDFILE>#;
'''

FX_STABILITY = '''<TITLE>FOLDX_runscript;
<JOBSTART>#;
    <PDBS>#;
    <BATCH>listNUM.txt;
    <COMMANDS>FOLDX_commandfile;
        <Stability>Stability.txt;
    <END>#;
    <OPTIONS>FOLDX_optionfile;
        <Temperature>TEMP;
        <R>#;
        <pH>7;
        <IonStrength>0.050;
        <water>-IGNORE;
        <metal>-IGNORE;
        <VdWDesign>2;
        <OutPDB>true;
        <pdb_hydrogens>false;
    <END>#;
<JOBEND>#;
<ENDFILE>#;
'''

FX_COMPLEX = '''<TITLE>FOLDX_runscript;
<JOBSTART>#;
    <PDBS>#;
    <BATCH>listNUM.txt;
    <COMMANDS>FOLDX_commandfile;
        <AnalyseComplex>#,CHAIN;
    <END>#;
    <OPTIONS>FOLDX_optionfile;
        <Temperature>TEMP;
        <R>#;
        <pH>7;
        <IonStrength>0.050;
        <water>-IGNORE;
        <metal>-IGNORE;
        <VdWDesign>2;
        <OutPDB>true;
        <pdb_hydrogens>false;
    <END>#;
<JOBEND>#;
<ENDFILE>#;
'''

#
# PARSE COMMAND LINE ARGUMENTS
#

parser = argparse.ArgumentParser(description = 'With -i and -n this script\
        takes an input PDB file and then generates scripts to perform\
        mutations and analyze folding (default) or binding free energies\
        (-g option) with foldx. With -f option only specified mutations will\
        be performed. Default is to calculate every possible amino acid\
        change. With -a this script will analyze the output of a run.' )
parser.add_argument('-a',
    metavar = 'DIRBASE',
    type = str,
    help = '(for post-processing analysis) base name for directories to\
            analyze' )
parser.add_argument('-i',
    metavar = 'PDBFILE',
    type = argparse.FileType('r'),
    help = '(requires -n) input PDB file' )
parser.add_argument('-n',
    metavar = 'NCORE',
    type = int,
    help = '(requires -i) number of cores to use' )
parser.add_argument('-f',
    metavar = 'MUTFILE',
    type = argparse.FileType('r'),
    help = 'input file containing column of mutants. format is\
            residue_from + chain + position + residue_to, e.g., "DA33L"\
            for single chain or "DA33L,DB33L" for two.' )
parser.add_argument('-g',
    metavar = 'CHAIN_GRP',
    type = str,
    help = '(required for complex) chain(s) to calculate for interaction\
            energy, e.g., -g A will\
            calculate interaction between chain A and everything else,\
            -g AD will calculate interaction between chains A+D and\
            everything else.' )
parser.add_argument('-m',
    metavar = 'CHAIN_MUT',
    type = str,
    help = '(required if using -g and not using -f) chain in complex to\
            mutate, e.g., -m A will mutate chain A,\
            -m BD will mutate chains B and D at the same time (dimer).' )
parser.add_argument('-b',
    metavar = 'BEGIN',
    type = int,
    default = 1,
    help = '(default=first) residue number to begin mutations' )
parser.add_argument('-e',
    metavar = 'END',
    type = int,
    default = 100000,
    help = '(default=last) residue number to end mutations' )
parser.add_argument('-r',
    metavar = 'NREPAIR',
    type = int,
    default = 0,
    help = '(default=0) number of times to used foldx RepairPDB on mutants\
            before analyzing' )
parser.add_argument('-t',
    metavar = 'TEMP',
    type = int,
    default = 298,
    help = '(default=298 K) temperature in Kelvin' )
parser.add_argument('-o',
    metavar = 'OUTBASE',
    type = str,
    default = 'fx_scan',
    help = '(default=fx_scan) base name for output directories' )
parser.add_argument('-p',
    action = 'store_true',
    help = '(default=false) generate PBS scripts for a computer cluster\
            instead of bash scripts' )

if len(sys.argv) == 1:  # Print help if no arguments are given.
    parser.print_help()
    sys.exit(1)

args = parser.parse_args()

# Argument dependencies
if not any( [args.a, args.i, args.n] ):
    sys.exit('error: must specify -a or -i/-n flags')
if args.i and not args.n or args.n and not args.i:
    sys.exit('error: must specify both -i and -n')

#
# FUNCTIONS
#

def analyze_foldx_dirs():
    '''
    Analyze directories containing output from previous foldx runs.
    '''

    deltag = []
    mutation = []
    complex_flag = None
    # Find all directories that match the given base name.
    for dir_name in sorted(glob.glob(args.a + '*')):
    
        # Check to see if we are calculating for a single chain or a complex.
        # Also, check to be sure we don't switch between the two.
        complex_base = glob.glob(dir_name + '/Summary_Analyse*wt.fxout')
        if complex_base:
            complex_base = complex_base[0].split('/')[1][:-6]
            if complex_flag is False:
                sys.exit('error: switched between single chain and complex')
            else:
                complex_flag = True
        elif os.path.exists(dir_name + '/Stability.txt'):
            if complex_flag is True:
                sys.exit('error: switched between single chain and complex')
            else:
                complex_flag = False
        else:
            sys.exit('error: foldx files not found: Stability.txt or' +\
                    ' Summary_*wt.fxout required')

        # Get list of mutations.
        num_mutations = 0
        with open(dir_name + '/individual_list.txt', 'r') as txt_file:
            for line in txt_file:
                num_mutations += 1
                mutation.append(line.rstrip()[:-1])

        # Calculate and store Delta G values.
        first_deltag = None
        if complex_flag:
            with open(dir_name + '/' + complex_base + '.fxout') as txt_file:
                for line in txt_file:
                    if line.startswith('Repair'):
                        first_deltag = float(line.split()[5])
            for imut in range(1, num_mutations+1):
                with open(dir_name + '/' + complex_base + '_' + str(imut) +\
                        '.fxout') as txt_file:
                    for line in txt_file:
                        if line.startswith('Repair'):
                            deltag.append(float(line.split()[5]) - first_deltag)
        else:
            with open(dir_name + '/Stability.txt', 'r') as txt_file:
                for line in txt_file:
                    if line.startswith('Repair'):
                        if not first_deltag:
                            first_deltag = float(line.split()[1])
                        else:
                            deltag.append(float(line.split()[1]) - first_deltag)

    if mutation and deltag:
        for c1, c2 in zip(mutation, deltag):
            if c1[0] is c1[-1]:
                if abs(c2) > SELF_MUT_DG_MAX:
                    sys.stderr.write('warning: large self-mutation: ' +\
                            '{:s}{:14.4e}\n'.format(c1, c2) )
            else:
                print('{:s} {:14.4e}'.format(c1, c2))
    else:
        sys.exit('error: no directories match base name: ' + args.a)

def process_pdb():
    '''
    Open and process PDB file, and generate list of all possible mutants.
    '''

    mutant_list = []
    first_chain = None
    for line in args.i:
        if line.startswith('ATOM'):
            resi_num = int(line[22:26].strip()) 
            atom_type = line[12:16].strip()
            chain = line[21]
            if not first_chain:
                first_chain = chain
            if chain is not first_chain:
                sys.exit('error: ' + args.i.name + ' contains multiple chains')
            if args.b <= resi_num <= args.e and atom_type == 'CA':
                for resi_to in RESI_DICT.itervalues():
                    resi_from = RESI_DICT[line[17:20]]
                    mutant_list.append(str(resi_from) + chain +\
                            str(resi_num) + str(resi_to) + ';')

    return mutant_list

def process_pdb_complex(chains_to_mutate):
    '''
    Open and process PDB file, and generate list of all possible mutants.
    This differs from process_pdb() because it analyzes a PDB complex.
    '''

    mutant_list = []
    for line in args.i:
        if line.startswith('ATOM'):
            resi_num = int(line[22:26].strip()) 
            atom_type = line[12:16].strip()
            chain = line[21]
            if args.b <= resi_num <= args.e and atom_type == 'CA' and\
                    chain in chains_to_mutate[0]:
                for resi_to in RESI_DICT.itervalues():
                    resi_from = RESI_DICT[line[17:20]]
                    mutant_line = ''
                    for ichain in chains_to_mutate:
                        mutant_line += str(resi_from) + ichain +\
                                str(resi_num) + str(resi_to) + ','
                    mutant_list.append(mutant_line[:-1] + ';')

    return mutant_list

def gen_fx_files(mutant_list, chains_to_mutate):
    '''
    Create directories and fill them with input files for foldx runs.
    '''

    # Calculate the number of mutants to process per directory.
    if len(mutant_list) % args.n == 0:
        mut_per_dir = len(mutant_list)/args.n
    else:
        mut_per_dir = int(len(mutant_list)/args.n) + 1 

    mut_start = 0
    mut_end = mut_per_dir
    run_max = args.r + 8
    for idir in range(1, args.n+1):
        dir_name = args.o + '{:02d}'.format(idir)
        # Make directory and copy required input PDB and database files.
        os.mkdir(dir_name)
        shutil.copy(args.i.name, dir_name + '/wt.pdb')
        shutil.copy(foldx + '/rotabase.txt', dir_name)
        # Build run.txt files for foldx.
        # Will perform 6 repairs -> build mutants -> args.r repairs.
        for irun in range(1, 7): # 6 repairs
            with open(dir_name + '/run' + str(irun) + '.txt', 'w') as txt_file:
                txt_file.write(FX_REPAIR.replace('NUM',\
                str(irun)).replace('TEMP', str(args.t)) )
        with open(dir_name + '/run7.txt', 'w') as txt_file: # build mutants
            txt_file.write(FX_BUILD.replace('NUM', '7').replace('TEMP',\
            str(args.t)) )
        for irun in range(8, run_max + 1): # repair mutants
            with open(dir_name + '/run' + str(irun) + '.txt', 'w') as txt_file:
                txt_file.write(FX_REPAIR.replace('NUM', str(irun)).replace(\
                'TEMP', str(args.t)) )
        if args.g: # calculate binding affinity for protein complex
            with open(dir_name + '/run' + str(run_max) + '.txt', 'w')\
                    as txt_file:
                txt_file.write(FX_COMPLEX.replace('CHAIN',\
                args.g).replace('NUM', str(run_max)).replace('TEMP',\
                str(args.t)) )
        else: # calculate folding stability for single protein chain
            with open(dir_name + '/run' + str(run_max) + '.txt', 'w')\
                    as txt_file:
                txt_file.write(FX_STABILITY.replace('NUM',\
                str(run_max)).replace('TEMP', str(args.t)) )
        # Build list.txt containing structures to evaluate.
        for ilist in range(1, 8):
            with open(dir_name + '/list' + str(ilist) +\
                    '.txt', 'w') as txt_file:
                jlist = ilist - 1
                prefix = 'RepairPDB_' * jlist
                txt_file.write(prefix + 'wt.pdb\n')
        for ilist in range(8, run_max + 1):
            with open(dir_name + '/list' + str(ilist) +\
                    '.txt', 'w') as txt_file:
                jlist = ilist - 2
                prefix = 'RepairPDB_' * jlist
                txt_file.write(prefix + 'wt.pdb\n')
                index = 0
                for imut in mutant_list[mut_start:mut_end]:
                    index += 1
                    txt_file.write(prefix + 'wt_' + str(index) + '.pdb\n')
        # Build list of all mutations needed run7.txt.
        with open(dir_name + '/individual_list.txt', 'w') as txt_file:
            if idir < args.n:
                for imut in mutant_list[mut_start:mut_end]:
                    txt_file.write(imut + '\n')
            else:
                for imut in mutant_list[mut_start:len(mutant_list)]:
                    txt_file.write(imut + '\n')
        # Build bash or PBS script files to run jobs.
        if args.p:
            with open(dir_name + '/job_foldx.pbs', 'w') as txt_file:
                txt_file.write('#!/bin/bash\n#PBS -N marty_foldx\n'\
                        '#PBS -l nodes=1:ppn=1\n#PBS -j oe\n#PBS -q reg\n\n'\
                        'cd $PBS_O_WORKDIR\n\n. /usr/modules/init/bash\n'
                        'module load foldx\n\n' )
                for irun in range(1, run_max + 1):
                    txt_file.write('foldx -runfile run' + str(irun) +\
                            '.txt >> foldx.log\n' )
            #print('job files generated for ' + dir_name)
            cmd = 'cd ' + dir_name + ' ; chmod +x job_foldx.pbs ; ' +\
                    'qsub job_foldx.pbs ; cd ../ ;'
            print 'starting job: ' + dir_name + '/job_foldx.pbs'
            subprocess.Popen(cmd, shell = True, executable = '/bin/bash',\
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            with open(dir_name + '/job_foldx.sh', 'w') as txt_file:
                txt_file.write('#!/bin/bash\n')
                for irun in range(1, run_max + 1):
                    txt_file.write(foldx + '/foldx -runfile run' + str(irun) +\
                            '.txt >> foldx.log\n')
            cmd = 'cd ' + dir_name + ' ; chmod +x job_foldx.sh ; ' +\
                    'job_foldx.sh &\\ ; cd ../ ;'
            print('starting job: ' + dir_name + '/job_foldx.sh')
            subprocess.Popen(cmd, shell = True, executable = '/bin/bash',\
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        mut_start += mut_per_dir
        mut_end += mut_per_dir

#
# MAIN PROGRAM
#

if __name__ == '__main__':

    # If outputing PBS scripts then update path to foldx directory.
    if args.p:
        foldx = pbs_foldx

    if args.a:
        analyze_foldx_dirs()
    
    else:
        # Check to make sure that directories don't already exist.
        test_dir = args.o + '01'
        if os.path.exists(test_dir):
            sys.exit('error: ' + test_dir + ' exists: will not overwrite')
        # Check to make sure that database file exists.
        if not os.path.exists(foldx + '/rotabase.txt'):
            sys.exit('error: ' + foldx + '/rotabase.txt not found')

        # No reason to exceed six RepairPDB calls because convergence obtained
        # around six.
        if args.r > 6:
            print('warning: option -r exeeds max value: changing to -r 6')
            args.r = 6
        
        # Process protein complex files.
        if args.g:
            if args.m:
                chains_to_mutate = list(args.m)
            else:
                chains_to_mutate = None
            if args.f:
                mutant_list = []
                for line in args.f:
                    mutant_list.append(line.rstrip() + ';')
            else:
                mutant_list = process_pdb_complex(chains_to_mutate)
        
        # Process single chain files.
        else:
            chains_to_mutate = None
            if args.f:
                mutant_list = []
                for line in args.f:
                    mutant_list.append(line.rstrip() + ';')
            else:
                mutant_list = process_pdb()
        
        gen_fx_files(mutant_list, chains_to_mutate)

    
        # Provided the code ran with problems then output command line arguments
        # and date/time to log file.
        log_file = os.path.basename(__file__)[:-3] + '.log'
        with open(log_file, 'a') as txt_file:
            now = datetime.datetime.now()
            now_date = str(now.month) + '/' + str(now.day) + '/' + str(now.year)
            now_time = str(now.hour) + ':' + str(now.minute) + ':' +\
                    str(now.second)
            txt_file.write(now_time + '  ' + now_date + '\n')
            txt_file.write(' '.join(sys.argv) + '\n\n')
